/*
			  Object Operation Dialog

	Used for the querying of copying, moving, linking, chowning,
	and changing permissions of objects.
 */

#ifndef OBJ_OP_DLG_H
#define OBJ_OP_DLG_H

#include <gtk/gtk.h>
#include "pulist.h"
#include "edv_types.h"
#include "libendeavour2-base/edv_vfs_obj.h"


typedef struct _edv_obj_op_dlg_struct	edv_obj_op_dlg_struct;
#define EDV_OBJ_OP_DLG(p)		((edv_obj_op_dlg_struct *)(p))


/*
 *	Operations:
 */
typedef enum {
	EDV_OBJ_OP_DLG_OP_MOVE,
	EDV_OBJ_OP_DLG_OP_COPY,
	EDV_OBJ_OP_DLG_OP_LINK,
	EDV_OBJ_OP_DLG_OP_CHMOD,		/* Change permissions */
	EDV_OBJ_OP_DLG_OP_CHOWN,		/* Change ownership */
	EDV_OBJ_OP_DLG_OP_CHTIME		/* Change time stamps */
} edv_obj_op_dlg_op;


/*
 *	Object Operation Dialog:
 */
struct _edv_obj_op_dlg_struct {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gboolean	processing;
	gint		freeze_count,
			busy_count;
	EDVCore *core;

	GtkWidget	*main_vbox,
			*icon_fixed,
			*icon_pm,
			*label,			/* For describing the source
						 * objects */

			/* Move, Copy, and Link Widgets */
			*target_entry,
			*target_browse_btn,

			/* Change Permissions Widgets */
			*target_chmod_parent,
			*target_ur_check,
			*target_uw_check,
			*target_ux_check,
			*target_gr_check,
			*target_gw_check,
			*target_gx_check,
			*target_or_check,
			*target_ow_check,
			*target_ox_check,
			*target_suid_check,
			*target_sgid_check,
			*target_sticky_check,

			/* Change Ownership Widgets */
			*target_chown_parent,
			*target_owner_entry,
			*target_owner_btn,
			*target_group_entry,
			*target_group_btn,

			/* Change Time Stamps Widgets */
			*target_time_parent,
			*target_atime_prompt,
			*target_mtime_prompt,
			*target_dtime_prompt,

			*opt_parent,
			*opt_recursive_check,
			*opt_dereference_links_check,

			*move_btn,
			*copy_btn,
			*link_btn,
			*chmod_btn,
			*chown_btn,
			*chtime_btn,

			*cancel_btn;

	/* Reference toplevel for transiency, shared */
	GtkWidget	*ref_toplevel;

	/* Operation that we are querying values for */
	edv_obj_op_dlg_op	op;

	/* Location type */
	EDVLocationType	location_type;

	/* Source list of objects, the data type is determined
	 * by the location_type, for example, if location_type
	 * EDV_LOCATION_TYPE_VFS then each data type in the list
	 * is an EDVVFSObject
	 */
	GList		*objs_list;

	/* Source directory (relative/current location) */
	gchar		*src_dir;

	/* Location of the archive (if the location_type is
	 * EDV_LOCATION_TYPE_ARCHIVE)
	 */
	gchar		*arch_path;

};


extern edv_obj_op_dlg_struct *EDVObjOpDlgNew(EDVCore *core);
extern void EDVObjOpDlgReset(edv_obj_op_dlg_struct *d);
extern void EDVObjOpDlgSetBusy(edv_obj_op_dlg_struct *d, const gboolean busy);
extern gboolean EDVObjOpDlgIsMapped(edv_obj_op_dlg_struct *d);
extern void EDVObjOpDlgMapValues(
	edv_obj_op_dlg_struct *d,
	const edv_obj_op_dlg_op op,
	const EDVLocationType location_type,
	GList *objs_list,
	const gchar *src_dir,
	GtkWidget *toplevel
);
extern void EDVObjOpDlgUnmap(edv_obj_op_dlg_struct *d);
extern void EDVObjOpDlgDelete(edv_obj_op_dlg_struct *d);


#endif	/* OBJ_OP_DLG_H */
